#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#define BLOCK_SIZE	512
#define	OFFSET_BSIZE	0x00b
#define	OFFSET_CSIZE	0x00d
#define	OFFSET_RSIZE	0x00e
#define	OFFSET_NBFAT	0x010
#define	OFFSET_COUNT	0x013
#define OFFSET_FSIZE	0x016

int main(int argc,char *argv[]){
if(argc!=2){
  fprintf(stderr,"Usage: %s <block device>\n",argv[0]);
  exit(EXIT_FAILURE);
  }
char *device=argv[1];
int b=open(device,O_RDONLY);
if(b<0){
  fprintf(stderr,"Cannot open '%s'\n",device);
  exit(EXIT_FAILURE);
  }
unsigned char block[BLOCK_SIZE];
int nb=read(b,block,BLOCK_SIZE);
if(nb!=BLOCK_SIZE){
  fprintf(stderr,"Cannot read boot sector\n");
  exit(EXIT_FAILURE);
  }
uint16_t bsize=*(uint16_t *)(block+OFFSET_BSIZE);
printf("Block size \t %d\n",bsize);
uint8_t csize=*(uint8_t *)(block+OFFSET_CSIZE);
printf("Cluster size \t %d\n",csize);
uint8_t rsize=*(uint8_t *)(block+OFFSET_RSIZE);
printf("Reserved size \t %d\n",rsize);
uint8_t nbfat=*(uint8_t *)(block+OFFSET_NBFAT);
printf("# of FAT \t %d\n",nbfat);
uint16_t fsize=*(uint16_t *)(block+OFFSET_FSIZE);
printf("Size of FAT \t %d\n",fsize);
uint16_t total=*(uint16_t *)(block+OFFSET_COUNT);
printf("# of blocks \t %d\n",total);
int i;
for(i=0;i<rsize-1;i++){
  nb=read(b,block,BLOCK_SIZE);
  if(nb!=BLOCK_SIZE){
    fprintf(stderr,"Cannot read reserved sector #%d\n",i);
    exit(EXIT_FAILURE);
    }
  }
int no=0;
uint8_t trio[3];
int allocated=0;
while(no<BLOCK_SIZE*fsize){
  if(no%BLOCK_SIZE==0){
    nb=read(b,block,BLOCK_SIZE);
    if(nb!=BLOCK_SIZE){
      fprintf(stderr,"Cannot read FAT sector #%d\n",no/BLOCK_SIZE);
      exit(EXIT_FAILURE);
      }
    }
  trio[no%3]=block[no%BLOCK_SIZE];
  if(no%3==2){ 
    uint16_t a=trio[0]|((trio[1]&0x0f)<<8);
    uint16_t b=trio[2]<<4|((trio[1]&0xf0)>>4);
    if(a>0) allocated++;
    if(b>0) allocated++;
    }
  no++;
  }
printf("Allocated %d/%d => %d%%\n",allocated*csize,total,allocated*csize*100/total);
close(b);
return 0;
}
